"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var error_1 = require("./error");
var fs_1 = require("./fs");
var path_1 = require("./path");
function getRollupDefaultPlugin(options) {
    return {
        name: 'Rollup Core',
        resolveId: createResolveId(options),
        load: function (id) {
            return fs_1.readFileSync(id, 'utf-8');
        },
        resolveDynamicImport: function (specifier, parentId) {
            if (typeof specifier === 'string' && !this.isExternal(specifier, parentId, false))
                return this.resolveId(specifier, parentId);
        },
        resolveAssetUrl: function (_a) {
            var relativeAssetPath = _a.relativeAssetPath, format = _a.format;
            return relativeUrlMechanisms[format](relativeAssetPath);
        },
        resolveImportMeta: function (prop, _a) {
            var chunkId = _a.chunkId, format = _a.format;
            var mechanism = importMetaMechanisms[format] && importMetaMechanisms[format](prop, chunkId);
            if (mechanism) {
                return mechanism;
            }
        }
    };
}
exports.getRollupDefaultPlugin = getRollupDefaultPlugin;
function findFile(file, preserveSymlinks) {
    try {
        var stats = fs_1.lstatSync(file);
        if (!preserveSymlinks && stats.isSymbolicLink())
            return findFile(fs_1.realpathSync(file), preserveSymlinks);
        if ((preserveSymlinks && stats.isSymbolicLink()) || stats.isFile()) {
            // check case
            var name_1 = path_1.basename(file);
            var files = fs_1.readdirSync(path_1.dirname(file));
            if (files.indexOf(name_1) !== -1)
                return file;
        }
    }
    catch (err) {
        // suppress
    }
}
function addJsExtensionIfNecessary(file, preserveSymlinks) {
    var found = findFile(file, preserveSymlinks);
    if (found)
        return found;
    found = findFile(file + '.mjs', preserveSymlinks);
    if (found)
        return found;
    found = findFile(file + '.js', preserveSymlinks);
    return found;
}
function createResolveId(options) {
    return function (importee, importer) {
        if (typeof process === 'undefined') {
            error_1.error({
                code: 'MISSING_PROCESS',
                message: "It looks like you're using Rollup in a non-Node.js environment. This means you must supply a plugin with custom resolveId and load functions",
                url: 'https://rollupjs.org/guide/en#a-simple-example'
            });
        }
        // external modules (non-entry modules that start with neither '.' or '/')
        // are skipped at this stage.
        if (importer !== undefined && !path_1.isAbsolute(importee) && importee[0] !== '.')
            return null;
        // `resolve` processes paths from right to left, prepending them until an
        // absolute path is created. Absolute importees therefore shortcircuit the
        // resolve call and require no special handing on our part.
        // See https://nodejs.org/api/path.html#path_path_resolve_paths
        return addJsExtensionIfNecessary(path_1.resolve(importer ? path_1.dirname(importer) : path_1.resolve(), importee), options.preserveSymlinks);
    };
}
var getResolveUrl = function (path, URL) {
    if (URL === void 0) { URL = 'URL'; }
    return "new " + URL + "(" + path + ").href";
};
var getUrlFromDocument = function (chunkId) {
    return "(document.currentScript && document.currentScript.src || new URL('" + chunkId + "', document.baseURI).href)";
};
var getGenericImportMetaMechanism = function (getUrl) { return function (prop, chunkId) {
    var urlMechanism = getUrl(chunkId);
    return prop === null ? "({ url: " + urlMechanism + " })" : prop === 'url' ? urlMechanism : 'undefined';
}; };
var importMetaMechanisms = {
    amd: getGenericImportMetaMechanism(function () { return getResolveUrl("module.uri, document.baseURI"); }),
    cjs: getGenericImportMetaMechanism(function (chunkId) {
        return "(typeof document === 'undefined' ? " + getResolveUrl("'file:' + __filename", "(require('u' + 'rl').URL)") + " : " + getUrlFromDocument(chunkId) + ")";
    }),
    iife: getGenericImportMetaMechanism(function (chunkId) { return getUrlFromDocument(chunkId); }),
    system: function (prop) { return (prop === null ? "module.meta" : "module.meta." + prop); },
    umd: getGenericImportMetaMechanism(function (chunkId) {
        return "(typeof document === 'undefined' ? " + getResolveUrl("'file:' + __filename", "(require('u' + 'rl').URL)") + " : " + getUrlFromDocument(chunkId) + ")";
    })
};
var getRelativeUrlFromDocument = function (relativePath) {
    return getResolveUrl("(document.currentScript && document.currentScript.src || document.baseURI) + '/../" + relativePath + "'");
};
var relativeUrlMechanisms = {
    amd: function (relativePath) { return getResolveUrl("module.uri + '/../" + relativePath + "', document.baseURI"); },
    cjs: function (relativePath) {
        return "(typeof document === 'undefined' ? " + getResolveUrl("'file:' + __dirname + '/" + relativePath + "'", "(require('u' + 'rl').URL)") + " : " + getRelativeUrlFromDocument(relativePath) + ")";
    },
    es: function (relativePath) { return getResolveUrl("'" + relativePath + "', import.meta.url"); },
    iife: function (relativePath) { return getRelativeUrlFromDocument(relativePath); },
    system: function (relativePath) { return getResolveUrl("'" + relativePath + "', module.meta.url"); },
    umd: function (relativePath) {
        return "(typeof document === 'undefined' ? " + getResolveUrl("'file:' + __dirname + '/" + relativePath + "'", "(require('u' + 'rl').URL)") + " : " + getRelativeUrlFromDocument(relativePath) + ")";
    }
};
